#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['putao', 'putao.backend']

package_data = \
{'': ['*'], 'putao': ['.ropeproject/*'], 'putao.backend': ['.ropeproject/*']}

install_requires = \
['chardet>=4.0.0', 'mido>=1.2.9', 'pydub>=0.25.1', 'pyparsing>=2.4.7']

setup(name='putao',
      version='0.1.0a0',
      description="葡萄 (putao, grape): Poor man's UTAU.",
      author='Ong Yong Xin',
      author_email='ongyongxin2020+github@gmail.com',
      url='https://github.com/ongyx/putao',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      python_requires='>=3.6',
     )
